@@ -24,29 +24,17 @@
 
 #include <gio/gio.h>
 
-G_GNUC_INTERNAL gboolean
-gssdp_socket_enable_loop         (GSocket       *socket,
-                                  gboolean       enable,
-                                  GError       **error);
-G_GNUC_INTERNAL gboolean
-gssdp_socket_set_ttl             (GSocket       *socket,
-                                  int            ttl,
-                                  GError       **error);
-G_GNUC_INTERNAL gboolean
-gssdp_socket_enable_broadcast    (GSocket       *socket,
-                                  gboolean       enable,
-                                  GError       **error);
 G_GNUC_INTERNAL gboolean
 gssdp_socket_mcast_interface_set (GSocket       *socket,
                                   GInetAddress  *iface_address,
                                   GError       **error);
 G_GNUC_INTERNAL gboolean
-gssdp_socket_mcast_group_join    (GSocket       *socket,
-                                  GInetAddress  *group,
-                                  GInetAddress  *iface,
-                                  GError       **error);
-G_GNUC_INTERNAL gboolean
 gssdp_socket_reuse_address       (GSocket *socket,
                                   gboolean enable,
                                   GError **error);
+
+G_GNUC_INTERNAL gboolean
+gssdp_socket_enable_info         (GSocket *socket,
+                                  gboolean enable,
+                                  GError **error);
 #endif
